import json
from pathlib import Path
import xml.etree.cElementTree as ET
from xml.dom import minidom

import numpy as np
from matplotlib import colors
from wholeslidedata.annotation.structures import Point, Polygon


def write_polygon(annos, coordinates, index, label_name, label_color):
    anno = ET.SubElement(annos, "Annotation")
    anno.set("Name", "Annotation " + str(index))
    anno.set("Type", "Polygon")
    anno.set("PartOfGroup", label_name)
    anno.set("Color", label_color)

    coords = ET.SubElement(anno, "Coordinates")
    coords_mem = []
    ridx = 0
    for _, r in enumerate(coordinates):
        x = int(r[0])
        y = int(r[1])
        if (x, y) in coords_mem:
            continue
        coords_mem.append((x, y))

        coord = ET.SubElement(coords, "Coordinate")
        coord.set("Order", str(ridx))
        coord.set("X", str(x))
        coord.set("Y", str(y))
        ridx += 1


def write_point(annos, coordinates, index, label_name, label_color):
    anno = ET.SubElement(annos, "Annotation")
    anno.set("Name", "Annotation " + str(index))
    anno.set("Type", "Dot")
    anno.set("PartOfGroup", label_name)
    anno.set("Color", label_color)

    coords = ET.SubElement(anno, "Coordinates")
    coords_mem = []
    ridx = 0
    for _, r in enumerate(coordinates):
        x = int(r[0])
        y = int(r[1])
        if (x, y) in coords_mem:
            continue
        coords_mem.append((x, y))

        coord = ET.SubElement(coords, "Coordinate")
        coord.set("Order", str(ridx))
        coord.set("X", str(x))
        coord.set("Y", str(y))
        ridx += 1


def write_point_set(annotations, output_path, label_name="tils", label_color="yellow"):
    # the root of the xml file.
    root = ET.Element("ASAP_Annotations")

    # writing each anno one by one.
    annos = ET.SubElement(root, "Annotations")

    # writing the last groups part
    anno_groups = ET.SubElement(root, "AnnotationGroups")

    index = 0
    anno = ET.SubElement(annos, "Annotation")
    anno.set("Name", "Annotation " + str(index))
    anno.set("Type", "PointSet")
    anno.set("PartOfGroup", label_name)
    anno.set("Color", label_color)

    coords = ET.SubElement(anno, "Coordinates")
    coords_mem = []
    ridx = 0
    for annotation in annotations:
        x, y = annotation.center
        if (x, y) in coords_mem:
            continue
        coords_mem.append((x, y))

        coord = ET.SubElement(coords, "Coordinate")
        coord.set("Order", str(ridx))
        coord.set("X", str(x))
        coord.set("Y", str(y))
        ridx += 1

    group = ET.SubElement(anno_groups, "Group")
    group.set("Name", label_name)
    group.set("PartOfGroup", "None")
    group.set("Color", label_color)
    attr = ET.SubElement(group, "Attributes")

    # writing to the xml file with indentation
    xmlstr = minidom.parseString(ET.tostring(root)).toprettyxml(indent="    ")
    with open(output_path, "w") as f:
        f.write(xmlstr)


def write_asap_annotation(annotations, output_path, scaling=1.0):
    # the root of the xml file.
    root = ET.Element("ASAP_Annotations")

    # writing each anno one by one.
    annos = ET.SubElement(root, "Annotations")

    # writing the last groups part
    anno_groups = ET.SubElement(root, "AnnotationGroups")

    labels = set()

    for annotation in annotations:
        label_name = annotation.label.name
        if annotation.label.weight is not None and annotation.label.weight > 0:
            label_name = label_name + "-weight=" + str(annotation.label.weight)
        label_color = annotation.label.color if annotation.label.color else "black"
        index = annotation.index
        if isinstance(annotation, Polygon):
            coordinates = annotation.coordinates / scaling
            write_polygon(annos, coordinates, index, label_name, label_color)
        elif isinstance(annotation, Point):
            coordinates = annotation.coordinates / scaling
            write_point(annos, [coordinates], index, label_name, label_color)
        else:
            raise ValueError("unsupported geometry", annotation)

        labels.add((label_name, label_color))

    for label, color in labels:
        group = ET.SubElement(anno_groups, "Group")
        group.set("Name", label)
        group.set("PartOfGroup", "None")
        group.set("Color", color)
        attr = ET.SubElement(group, "Attributes")

    # writing to the xml file with indentation
    xmlstr = minidom.parseString(ET.tostring(root)).toprettyxml(indent="    ")
    with open(output_path, "w") as f:
        f.write(xmlstr)


def write_qupath_annotation(wsa: Path, output_path=None, upsampling_ratio=1.0):
    # the root of the xml file
    if output_path is None:
        output_path = wsa.path.parent / (wsa.path.stem + ".json")

    labels = set()
    json_content = {}

    for annotation in wsa.annotations:
        label_name = annotation.label.name
        label_color = annotation.label.color if annotation.label.color else "black"
        label_rgb_color = colors.to_rgb(label_color)
        index = annotation.index
        polygon = annotation.buffer(1)
        coordinates = polygon.exterior.coords
        coordinates = np.array(coordinates) * upsampling_ratio
        coordinates = coordinates.astype(int).tolist()
        annotation_content = {
            "name": f"Annotation {index}",
            "color": label_rgb_color,
            "vertices": list(coordinates),
        }
        json_content.setdefault(label_name, []).append(annotation_content)

    with open(output_path, "w") as outfile:
        json.dump(json_content, outfile)
